#include <string.h>
#include <strings.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h> // debuginfo only
#include <sys/stat.h>

#include <adwin-macros.h>
#include <libadwin/limits.h>
#include "configfile.h"

adwin_dev_t device_list[MAX_DEVICES];
int configured_devices = 0;

extern int configFileparse(void);
extern FILE *configFilein, *configFileout;

int findDevice(int device_no)
{
    PRINTDEBUG(DBG_CONFIGFILE, ("searching for device %d...", device_no));
    int i;
    for (i = 0; i < configured_devices; i++)
        if (device_list[i].dev_name == device_no) {
            PRINTDEBUG(DBG_CONFIGFILE, ("found\n"));
            return i;
        }
    PRINTDEBUG(DBG_CONFIGFILE, ("NOT found\n"));
    return -1; // not found
}

adwin_dev_t getDevice(int device_index)
{
    return device_list[device_index];
}


void clearDevice(adwin_dev_t *dev)
{
    dev->dev_uid = 0;                              // root
    dev->dev_gid = 0;                              // root
    dev->dev_mode = 0666;                          // rw access for all
    dev->dev_name = 0x150;                         // default DeviceNo
    strcpy(dev->dev_desc, "");                     // no description
    dev->dev_type = -1;                            // invalid, test later
    dev->dev_link_base_addr = -1;        // invalid, test if type == link
    dev->dev_usb_serial = -1;            // invalid, test it if type == usb
    dev->dev_net_host_link = 0;                    // unused by default
    strcpy(dev->dev_net_host_password, "");        // no password (default)
    dev->dev_net_package_count = 5;                // default
    dev->dev_net_timeout = 1000;                   // default (1000ms)
    dev->dev_net_server.sin_port   = htons(6543);  // default
    dev->dev_net_server.sin_family = AF_INET;      // DON'T CHANGE
    bzero(&(dev->dev_net_server.sin_zero), 8);     // DON'T CHANGE
    inet_aton("0.0.0.0",                           // default, test later
              &dev->dev_net_server.sin_addr);
    // todo:
    // int dev_processor;
}

void reloadConfigFile()
{
    PRINTDEBUG(DBG_CONFIGFILE, ("reloading configuration...\n"));
    // don't read the config file again, unless reloadConfig is called
    // explicitly
    if (! existsConfigFile()) {
        fprintf(stderr, "configfile not found\n");
        return;
    }

    configured_devices = 0;
    configFilein = fopen(CONFIG_FILE, "r");
    if (configFilein == NULL) {
        perror("open config file:");
        return;
    }
    int result = configFileparse();
    fclose(configFilein);

    if (result) {
        fprintf(stderr, "Parse error in configuration file\n");
        exit(1);
    }

    PRINTDEBUG(DBG_CONFIGFILE, ("Parse returned %d, devices found: %d\n",
                                result, configured_devices));
    PRINTDEBUG(DBG_CONFIGFILE, ("finished reloading configuration\n"));
}

void printDeviceConfig(FILE *fd, adwin_dev_t dev)
{
    char buf[INET_ADDRSTRLEN];

    // todo: usb/link support
    fprintf(fd, "[0x%x]\n",     dev.dev_name);
    fprintf(fd, "\tUID = %d\n", dev.dev_uid);
    fprintf(fd, "\tGID = %d\n", dev.dev_gid);
    fprintf(fd, "\tMODE = 0%o\n", dev.dev_mode);
    fprintf(fd, "\tTYPE = ");
    switch (dev.dev_type) {
        case DC_ETHERNET:
            fprintf(fd, "net\n");
            fprintf(fd, "\tDESC = \"%s\"\n", dev.dev_desc);
            fprintf(fd, "\tHOSTLINK = 0x%x\n", dev.dev_net_host_link);
            fprintf(fd, "\tPW = \"%s\"\n",
                    dev.dev_net_host_password);
            fprintf(fd, "\tCOUNT = %d\n", dev.dev_net_package_count);
            fprintf(fd, "\tTIMEOUT = %d\n", dev.dev_net_timeout);
            fprintf(fd, "\tPORT = %d\n", ntohs(dev.dev_net_server.sin_port));
            inet_ntop(AF_INET, (void *)&dev.dev_net_server.sin_addr,
                      buf, INET_ADDRSTRLEN);
            fprintf(fd, "\tIP = %s\n", buf);
            break;
        default:
            fprintf(fd, "unknown\n");
            break;
    }
}

void printConfig(FILE *fd)
{
    int i;
    for (i = 0; i < configured_devices; i++)
        printDeviceConfig(fd, device_list[i]);
}

void createBackupCF()
{
    if (existsConfigFile()) {
        // move current config file to backup
        char backup[255];
        sprintf(backup, "%s.backup", CONFIG_FILE);
        if (rename(CONFIG_FILE, backup))
            perror("creating backup file");
    } else {
        PRINTDEBUG(DBG_CONFIGFILE,
                   ("not creating backup file: no configfile found\n"));
    }
}

void addDeviceCF(adwin_dev_t dev)
{
    PRINTDEBUG(DBG_CONFIGFILE, ("Adding device to config file\n"));

    reloadConfigFile();

    int index = findDevice(dev.dev_name);
    if (index > -1) {
        fprintf(stderr, "error: device with id 0x%x exists\n",
                dev.dev_name);
        exit(1);
    }

    createBackupCF();

    device_list[configured_devices] = dev;
    configured_devices++;

    // open the new config file
    FILE *fd;
    fd = fopen(CONFIG_FILE, "a");
    if (! fd) {
        perror("opening configuration file");
        exit(1);
    }

    printConfig(fd);
    fclose(fd);
}

void deleteDeviceCF(int device)
{
    PRINTDEBUG(DBG_CONFIGFILE, ("Deleting device from config file\n"));
    reloadConfigFile();
    int index = findDevice(device);
    if (index == -1) {
        fprintf(stderr, "error: no such device\n");
        exit(1);
    }

    if (index < configured_devices - 1) {
        // device to be deleted is not last device
        // copy last device to deleted device
        device_list[index] = device_list[configured_devices - 1];
    }
    configured_devices--;

    createBackupCF();

    // open the new config file
    FILE *fd;
    fd = fopen(CONFIG_FILE, "w");
    if (! fd) {
        perror("creating configuration file");
        exit(1);
    }

    printConfig(fd);
    fclose(fd);
}

int existsConfigFile()
{
    struct stat filestat;
    if (stat(CONFIG_FILE, &filestat)) {
        PRINTDEBUG(DBG_CONFIGFILE, ("config file does not exist!\n"));
        return 0;
    }
    return 1;
}

int getAdwindir(char *adwindir)
{
    FILE *configFile;

    configFile = fopen(ADWINDIR_FILE, "r");
    if (configFile == NULL) {
        PRINTDEBUG(DBG_CONFIGFILE,
                   ("Unable to open %s\n", ADWINDIR_FILE));
        return 0;
    }

    if (fscanf(configFile, "%254s", adwindir)  == EOF) {
        PRINTDEBUG(DBG_CONFIGFILE, ("%s is empty\n", ADWINDIR_FILE));
        return 0;
    }
    fclose(configFile);
    return 1;
}

int getDeviceCount()
{
    return configured_devices;
}
